from typing import List, Union

from graphql import GraphQLSchema, build_schema

from .types import Bindable


def make_executable_schema(
    type_defs: Union[str, List[str]],
    resolvers: Union[Bindable, List[Bindable], None] = None,
) -> GraphQLSchema:
    if isinstance(type_defs, list):
        type_defs = join_type_defs(type_defs)

    schema = build_schema(type_defs)

    if isinstance(resolvers, list):
        for type_resolvers in resolvers:
            type_resolvers.bind_to_schema(schema)
    elif resolvers:
        resolvers.bind_to_schema(schema)

    return schema


def join_type_defs(type_defs: List[str]) -> str:
    return "\n\n".join(t.strip() for t in type_defs)



from typing import Optional

from graphql.type import GraphQLScalarType, GraphQLSchema

from .types import Bindable, ScalarOperation


class Scalar(Bindable):
    _serialize: Optional[ScalarOperation]
    _parse_value: Optional[ScalarOperation]
    _parse_literal: Optional[ScalarOperation]

    def __init__(self, name: str) -> None:
        self.name = name
        self._serialize = None
        self._parse_value = None
        self._parse_literal = None

    def serializer(self, f: ScalarOperation) -> ScalarOperation:
        self._serialize = f
        return f

    def value_parser(self, f: ScalarOperation) -> ScalarOperation:
        self._parse_value = f
        return f

    def literal_parser(self, f: ScalarOperation) -> ScalarOperation:
        self._parse_literal = f
        return f

    def bind_to_schema(self, schema: GraphQLSchema) -> None:
        graphql_type = schema.type_map.get(self.name)
        self.validate_graphql_type(graphql_type)

        if self._serialize:
            graphql_type.serialize = self._serialize
        if self._parse_value:
            graphql_type.parse_value = self._parse_value
        if self._parse_literal:
            graphql_type.parse_literal = self._parse_literal

    def validate_graphql_type(self, graphql_type) -> None:
        if not graphql_type:
            raise ValueError("Scalar %s is not defined in the schema" % self.name)
        if not isinstance(graphql_type, GraphQLScalarType):
            raise ValueError(
                "%s is defined in the schema, but it is instance of %s (expected %s)"
                % (self.name, type(graphql_type).__name__, GraphQLScalarType.__name__)
            )



from typing import Any, Callable
from typing_extensions import Protocol

from graphql.type import GraphQLSchema


class Bindable(Protocol):
    def bind_to_schema(self, schema: GraphQLSchema) -> None:
        pass  # pragma: no cover


# Note: this should be [Any, GraphQLResolveInfo, **kwargs],
# but this is not achieveable with python types yet:
# https://github.com/mirumee/ariadne/pull/79
Resolver = Callable[..., Any]

ScalarOperation = Callable[[Any], Any]



from typing import List, Union

from .types import Bindable
from .wsgi_middleware import GraphQLMiddleware


def start_simple_server(
    type_defs: Union[str, List[str]],
    resolvers: Union[Bindable, List[Bindable], None] = None,
    host: str = "127.0.0.1",
    port: int = 8888,
) -> None:
    try:
        print("Simple GraphQL server is running on the http://%s:%s" % (host, port))
        graphql_server = GraphQLMiddleware.make_simple_server(
            type_defs, resolvers, host, port
        )
        graphql_server.serve_forever()
    except KeyboardInterrupt:
        pass



from .executable_schema import make_executable_schema
from .load_schema import load_schema_from_path
from .resolvers import (
    FallbackResolversSetter,
    ResolverMap,
    SnakeCaseFallbackResolversSetter,
    default_resolver,
    fallback_resolvers,
    resolve_to,
    snake_case_fallback_resolvers,
)
from .scalars import Scalar
from .simple_server import start_simple_server
from .utils import convert_camel_case_to_snake, gql
from .wsgi_middleware import GraphQLMiddleware


__all__ = [
    "FallbackResolversSetter",
    "GraphQLMiddleware",
    "ResolverMap",
    "Scalar",
    "SnakeCaseFallbackResolversSetter",
    "convert_camel_case_to_snake",
    "default_resolver",
    "fallback_resolvers",
    "gql",
    "load_schema_from_path",
    "make_executable_schema",
    "resolve_to",
    "snake_case_fallback_resolvers",
    "start_simple_server",
]



from graphql import parse


def convert_camel_case_to_snake(graphql_name: str) -> str:
    python_name = ""
    for i, c in enumerate(graphql_name.lower()):
        if i and c != graphql_name[i]:
            python_name += "_"
        python_name += c
    return python_name


def gql(value: str) -> str:
    parse(value)
    return value



import json
from typing import Any, Callable, List, Optional, Union
from wsgiref import simple_server

from graphql import GraphQLError, format_error, graphql_sync
from graphql.execution import ExecutionResult

from .constants import (
    CONTENT_TYPE_JSON,
    CONTENT_TYPE_TEXT_HTML,
    CONTENT_TYPE_TEXT_PLAIN,
    DATA_TYPE_JSON,
    HTTP_STATUS_200_OK,
    HTTP_STATUS_400_BAD_REQUEST,
    PLAYGROUND_HTML,
)
from .exceptions import HttpBadRequestError, HttpError, HttpMethodNotAllowedError
from .executable_schema import make_executable_schema
from .types import Bindable


class GraphQLMiddleware:
    def __init__(
        self,
        app: Optional[Callable],
        type_defs: Union[str, List[str]],
        resolvers: Union[Bindable, List[Bindable], None] = None,
        path: str = "/graphql/",
    ) -> None:
        self.app = app
        self.path = path
        self.schema = make_executable_schema(type_defs, resolvers)

        if not path:
            raise ValueError("path keyword argument can't be empty")

        if app is not None and not callable(app):
            raise TypeError("first argument must be a callable or None")

        if not app and path != "/":
            raise TypeError(
                "can't set custom path on WSGI middleware without providing "
                "application callable as first argument"
            )

        if app and path == "/":
            raise ValueError(
                "WSGI middleware can't use root path together with "
                "application callable"
            )

    def __call__(self, environ: dict, start_response: Callable) -> List[bytes]:
        if self.app and not environ["PATH_INFO"].startswith(self.path):
            return self.app(environ, start_response)

        try:
            return self.handle_request(environ, start_response)
        except GraphQLError as error:
            return self.handle_graphql_error(error, start_response)
        except HttpError as error:
            return self.handle_http_error(error, start_response)

    def handle_graphql_error(
        self, error: GraphQLError, start_response: Callable
    ) -> List[bytes]:
        start_response(
            HTTP_STATUS_400_BAD_REQUEST, [("Content-Type", CONTENT_TYPE_JSON)]
        )
        error_json = {"errors": [{"message": error.message}]}
        return [json.dumps(error_json).encode("utf-8")]

    def handle_http_error(
        self, error: HttpError, start_response: Callable
    ) -> List[bytes]:
        start_response(error.status, [("Content-Type", CONTENT_TYPE_TEXT_PLAIN)])
        response_body = error.message or error.status
        return [str(response_body).encode("utf-8")]

    def handle_request(self, environ: dict, start_response: Callable) -> List[bytes]:
        if environ["REQUEST_METHOD"] == "GET":
            return self.handle_get(start_response)
        if environ["REQUEST_METHOD"] == "POST":
            return self.handle_post(environ, start_response)
        raise HttpMethodNotAllowedError()

    def handle_get(self, start_response) -> List[bytes]:
        start_response(HTTP_STATUS_200_OK, [("Content-Type", CONTENT_TYPE_TEXT_HTML)])
        return [PLAYGROUND_HTML.encode("utf-8")]

    def handle_post(self, environ: dict, start_response: Callable) -> List[bytes]:
        data = self.get_request_data(environ)
        self.validate_query(data)
        result = self.execute_query(environ, data)
        return self.return_response_from_result(start_response, result)

    def get_request_data(self, environ: dict) -> dict:
        if environ.get("CONTENT_TYPE") != DATA_TYPE_JSON:
            raise HttpBadRequestError(
                "Posted content must be of type {}".format(DATA_TYPE_JSON)
            )

        request_content_length = self.get_request_content_length(environ)
        request_body = self.get_request_body(environ, request_content_length)

        data = self.parse_request_body(request_body)
        if not isinstance(data, dict):
            raise GraphQLError("Valid request body should be a JSON object")

        return data

    def get_request_content_length(self, environ: dict) -> int:
        try:
            content_length = int(environ.get("CONTENT_LENGTH", 0))
            if content_length < 1:
                raise HttpBadRequestError(
                    "Content length header is missing or incorrect"
                )
            return content_length
        except (TypeError, ValueError):
            raise HttpBadRequestError("Content length header is missing or incorrect")

    def get_request_body(self, environ: dict, content_length: int) -> bytes:
        if not environ.get("wsgi.input"):
            raise HttpBadRequestError("Request body cannot be empty")
        request_body = environ["wsgi.input"].read(content_length)
        if not request_body:
            raise HttpBadRequestError("Request body cannot be empty")
        return request_body

    def parse_request_body(self, request_body: bytes) -> Any:
        try:
            return json.loads(request_body)
        except ValueError:
            raise HttpBadRequestError("Request body is not a valid JSON")

    def validate_query(self, data: dict) -> None:
        self.validate_query_body(data.get("query"))
        self.validate_variables(data.get("variables"))
        self.validate_operation_name(data.get("operationName"))

    def validate_query_body(self, query) -> None:
        if not query or not isinstance(query, str):
            raise GraphQLError("The query must be a string.")

    def validate_variables(self, variables) -> None:
        if variables is not None and not isinstance(variables, dict):
            raise GraphQLError("Query variables must be a null or an object.")

    def validate_operation_name(self, operation_name) -> None:
        if operation_name is not None and not isinstance(operation_name, str):
            raise GraphQLError('"%s" is not a valid operation name.' % operation_name)

    def execute_query(self, environ: dict, data: dict) -> ExecutionResult:
        return graphql_sync(
            self.schema,
            data.get("query"),
            root_value=self.get_query_root(environ, data),
            context_value=self.get_query_context(environ, data),
            variable_values=data.get("variables"),
            operation_name=data.get("operationName"),
        )

    def get_query_root(
        self, environ: dict, request_data: dict  # pylint: disable=unused-argument
    ) -> Any:
        """Override this method in inheriting class to create query root."""
        return None

    def get_query_context(
        self, environ: dict, request_data: dict  # pylint: disable=unused-argument
    ) -> Any:
        """Override this method in inheriting class to create query context."""
        return {"environ": environ}

    def return_response_from_result(
        self, start_response: Callable, result: ExecutionResult
    ) -> List[bytes]:
        response = {"data": result.data}
        if result.errors:
            response["errors"] = [format_error(e) for e in result.errors]

        start_response(HTTP_STATUS_200_OK, [("Content-Type", CONTENT_TYPE_JSON)])
        return [json.dumps(response).encode("utf-8")]

    @classmethod
    def make_simple_server(
        cls,
        type_defs: Union[str, List[str]],
        resolvers: Union[Bindable, List[Bindable], None] = None,
        host: str = "127.0.0.1",
        port: int = 8888,
    ):
        wsgi_app = cls(None, type_defs, resolvers, path="/")
        return simple_server.make_server(host, port, wsgi_app)



from typing import Any, Callable, Dict, overload

from graphql.type import (
    GraphQLField,
    GraphQLObjectType,
    GraphQLResolveInfo,
    GraphQLSchema,
)

from .types import Bindable, Resolver
from .utils import convert_camel_case_to_snake


class ResolverMap(Bindable):
    _resolvers: Dict[str, Resolver]

    def __init__(self, name: str) -> None:
        self.name = name
        self._resolvers = {}

    @overload
    def field(self, name: str) -> Callable[[Resolver], Resolver]:
        pass  # pragma: no cover

    @overload
    def field(  # pylint: disable=function-redefined
        self, name: str, *, resolver: Resolver
    ) -> Resolver:  # pylint: disable=function-redefined
        pass  # pragma: no cover

    def field(self, name, *, resolver=None):  # pylint: disable=function-redefined
        if not resolver:
            return self.create_register_resolver(name)
        self._resolvers[name] = resolver
        return resolver

    def create_register_resolver(self, name: str) -> Callable[[Resolver], Resolver]:
        def register_resolver(f: Resolver) -> Resolver:
            self._resolvers[name] = f
            return f

        return register_resolver

    def alias(self, name: str, to: str) -> None:
        self._resolvers[name] = resolve_to(to)

    def bind_to_schema(self, schema: GraphQLSchema) -> None:
        graphql_type = schema.type_map.get(self.name)
        self.validate_graphql_type(graphql_type)

        for field, resolver in self._resolvers.items():
            if field not in graphql_type.fields:
                raise ValueError(
                    "Field %s is not defined on type %s" % (field, self.name)
                )

            graphql_type.fields[field].resolve = resolver

    def validate_graphql_type(self, graphql_type: str) -> None:
        if not graphql_type:
            raise ValueError("Type %s is not defined in the schema" % self.name)
        if not isinstance(graphql_type, GraphQLObjectType):
            raise ValueError(
                "%s is defined in the schema, but it is instance of %s (expected %s)"
                % (self.name, type(graphql_type).__name__, GraphQLObjectType.__name__)
            )


class FallbackResolversSetter(Bindable):
    def bind_to_schema(self, schema: GraphQLSchema) -> None:
        for type_object in schema.type_map.values():
            if isinstance(type_object, GraphQLObjectType):
                self.add_resolvers_to_object_fields(type_object)

    def add_resolvers_to_object_fields(self, type_object) -> None:
        for field_name, field_object in type_object.fields.items():
            self.add_resolver_to_field(field_name, field_object)

    def add_resolver_to_field(self, _: str, field_object: GraphQLField) -> None:
        if field_object.resolve is None:
            field_object.resolve = default_resolver


class SnakeCaseFallbackResolversSetter(FallbackResolversSetter):
    def add_resolver_to_field(
        self, field_name: str, field_object: GraphQLField
    ) -> None:
        if field_object.resolve is None:
            field_name = convert_camel_case_to_snake(field_name)
            field_object.resolve = resolve_to(field_name)


fallback_resolvers = FallbackResolversSetter()
snake_case_fallback_resolvers = SnakeCaseFallbackResolversSetter()


def resolve_parent_field(parent: Any, name: str) -> Any:
    if isinstance(parent, dict):
        return parent.get(name)
    return getattr(parent, name, None)


def default_resolver(parent: Any, info: GraphQLResolveInfo) -> Resolver:
    return resolve_parent_field(parent, info.field_name)


def resolve_to(name: str) -> Resolver:
    def resolver(parent: Any, *_) -> Any:
        return resolve_parent_field(parent, name)

    return resolver
